;;;;; helper routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; get_data_block_start_and_stop_numbers
;
; After opening a data archive file, this routine can be used to quickly
; determine which data structs in the file are the first and last structs
; in the file which meet the time interval criteria. 

function get_data_block_start_and_stop_numbers, day1996, dataFilePath,	       $
		structToUse, getStampsAfter, getStampsUpTo, structsInFileCount
    HEADER_LENGTH = 32
    assocOpenFlag = 0

    structName = strlowcase(tag_names(structToUse, /STRUCTURE_NAME))
    dataPeriod = get_data_type_period_from_struct_name(structName)

    secondsSince1996AtDayStart = get_seconds_since_1996_from_day_1996(day1996)
    secondsIntoDay = getStampsAfter - secondsSince1996AtDayStart

    if (secondsIntoDay LE 0) then startNumber = 0			       $
    else begin
	get_lun, lun
	openr, lun, dataFilePath
	fileAssoc = assoc(lun, structToUse, HEADER_LENGTH)
	assocOpenFlag = 1

	if (dataPeriod GT 0) then					       $
			dataBlockNumberInFile = secondsIntoDay / dataPeriod    $
	else dataBlockNumberInFile = 0

	    ; step forward through file from here to get right data block
	while (dataBlockNumberInFile LT structsInFileCount) do begin
	    struct = fileAssoc(dataBlockNumberInFile)
	    timeStamp = get_time_stamps_of_structs(struct)
	    if (timeStamp GT getStampsAfter) then goto, OUT1
	    dataBlockNumberInFile = dataBlockNumberInFile + 1
	endwhile

	OUT1:
	if (dataBlockNumberInFile EQ structsInFileCount) then		       $
							return, [ -1, -1 ]     $
	else startNumber = dataBlockNumberInFile
    endelse

    secondsIntoDay = getStampsUpTo - secondsSince1996AtDayStart
    if (secondsIntoDay GE 86399) then stopNumber = structsInFileCount - 1      $
    else begin
	if (NOT assocOpenFlag) then begin
	    get_lun, lun
	    openr, lun, dataFilePath
	    fileAssoc = assoc(lun, structToUse, HEADER_LENGTH)
	    assocOpenFlag = 1
	endif

	if (dataPeriod GT 0) then					       $
			dataBlockNumberInFile = secondsIntoDay / dataPeriod    $
	else dataBlockNumberInFile = startNumber

	while (dataBlockNumberInFile LT structsInFileCount) do begin
	    struct = fileAssoc(dataBlockNumberInFile)
	    timeStamp = get_time_stamps_of_structs(struct)
	    if (timeStamp GT getStampsUpTo) then goto, OUT2
	    dataBlockNumberInFile = dataBlockNumberInFile + 1
	endwhile

	OUT2:
	    ; we went 1 block past the last struct in the time interval
	stopNumber = dataBlockNumberInFile - 1
    endelse

    if (assocOpenFlag) then begin
	close, lun
	free_lun, lun
    endif

    return, [startNumber, stopNumber]
end

; get_subset_of_days_structs
;
; getSubsetOfDaysStructs is the real data retrieval engine. All calls to
; retrieve data from the archive go through this routine at some point. 
; 
; getSubsetOfDaysStructs will look at 1 file pointed at by path, and return
; an array of structs of type structName which have time
; stamps greater than getStampsAfter and less than or equal to
; getStampsUpTo. It does not care about times of validity of data structs,
; but only what the start time stamp of a struct is. Times of validity must be
; dealt with before the call to getSubsetOfDaysStructs in order to just hand
; it the values getStampsAfter and getStampsUpTo.

function get_subset_of_days_structs, day1996, structToUse, 		       $
					getStampsAfter, getStampsUpTo
    HEADER_LENGTH = 32
    structName = get_struct_name_from_struct(structToUse)

    date = get_date_from_day_1996(day1996)
    dataFilePath = generate_data_file_path(date, structName)

	; what if data file does not exist?
    if (file_exists(dataFilePath) EQ 0) then begin
	; print, 'FLAG: ', dataFilePath, ' does not exist.  No data that day?'
	structArray = replicate(structToUse, 1)
	return, structArray
    endif

    structsInFileCount = get_files_struct_count(dataFilePath, structToUse)
    if (structsInFileCount EQ 0) then begin
	structArray = replicate(structToUse, 1)
	return, structArray
    endif

    startAndStopIndicies = get_data_block_start_and_stop_numbers(day1996,      $
				dataFilePath, structToUse, getStampsAfter,     $
				getStampsUpTo, structsInFileCount)
    startIndex = startAndStopIndicies[0]
    stopIndex = startAndStopIndicies[1]
    numberOfStructsToRead = stopIndex - startIndex + 1

    if ((startIndex EQ -1) OR (numberOfStructsToRead LE 0)) then begin
	structArray = replicate(structToUse, 1)
	return, structArray
    endif

    structSize = get_struct_size(structToUse)
    byteOffset = HEADER_LENGTH + (startIndex * structSize)

    get_lun, lun
    openr, lun, dataFilePath

	; an assoc is used instead of readu because all structs are
	; padded out to an int address when written if they aren't
	; an integral number of int lenghts long.  assoc understands
	; this and takes care of it automatically.

    fileAssoc =								       $
	assoc(lun, replicate(structToUse, numberOfStructsToRead), byteOffset)
    structArray = fileAssoc(0)

    close, lun
    free_lun, lun

    stamps = get_time_stamps_of_structs(structArray)
    indiciesOfNonZeroStamps = where(stamps NE 0, count)
    if (count GT 0) then structArray = structArray(indiciesOfNonZeroStamps)

    structArray = [structArray, structToUse]	; append terminal struct
    return, structArray
end

; zero_last_structs_if_valid_at_stop_time
;
; Before get_structs returns the array of data structs a last check on the
; time of validity is done on the last structs to be sure they were not valid
; at the stop time in the original call.  If some are valid, this routine
; chops them off and returns true.  If it is not valid at stop time, nothing
; is done to the array and false is returned.

function zero_last_structs_if_valid_at_stop_time, structs, stopTime, structToUse
    lateStructsFound = 0

	; starting point at end of array
    indexOfLastGoodStruct = n_elements(structs) - 1

	; back up past all zeroed structs
    while (get_time_stamps_of_structs(structs[indexOfLastGoodStruct]) EQ 0) $
									do begin
	indexOfLastGoodStruct = indexOfLastGoodStruct - 1
	if (indexOfLastGoodStruct LT 0) then begin
	    structs = structToUse	; return with emptied structs array
	    return, 0
	endif
    endwhile

	; back up past any structs whose end time is >= stopTime
    lastSecondOfStructsValidity =					       $
		get_stop_times_of_structs(structs[indexOfLastGoodStruct])
    while (lastSecondOfStructsValidity GE stopTime) do begin
	indexOfLastGoodStruct = indexOfLastGoodStruct - 1
	if (indexOfLastGoodStruct LT 0) then begin
	    structs = structToUse	; return with emptied structs array
	    return, lateStructsFound
	endif
	lastSecondOfStructsValidity =					       $
		get_stop_times_of_structs(structs[indexOfLastGoodStruct])
	lateStructsFound = lateStructsFound + 1
    endwhile
	
	; chop off all offending data structs and add terminal zeroed struct
    structs = [structs[0:indexOfLastGoodStruct], structToUse]

    return, lateStructsFound
end

; get_begin_end_secs_of_struct_prior_to
;
; This routine is neccessary for handling those data structures which do
; not have a constant time period of validity (e.g. diagnostic events). When
; the user asks for such a data struct associated with a time
; secondsSince1996, we must step back through the data file from
; secondsSince1996 until we find a non-zero struct. Once found, we
; compare secondsSince1996 with the start and stop times of the struct to
; see if secondsSince1996 took place while this struct was valid. This
; stepping back may proceed back across a day boundary, but it is
; assumed that on not finding a non-zero struct throughout the entire
; previous day, no such struct will be found which is valid for the time
; (secondsSince1996) requested. 

function get_begin_end_secs_of_struct_prior_to, secondsSince1996, structToUse
    HEADER_LENGTH = 32
    beginAndEndTimeTags = lonarr(2)
    structName = get_struct_name_from_struct(structToUse)

    ct = 0	; command tables
    de = 0	; diagnostic events
    if (strpos(structName, 'commandtable') GE 0) then ct = 1		       $
    else if (strpos(structName, 'diagnosticevent') GE 0) then de = 1	       $
    else begin
	    ; no other structures should be using this routine
	print, 'Error: calling the get_begin_end_secs_of_struct_prior_to'
	print, 'for the ', structName, ' structure is not valid.'
	print, 'Stopping.'
	stop
    endelse

    day = get_day_1996_from_seconds_since_1996(secondsSince1996)
    date = get_date_from_day_1996(day)
    dayStartSecond = cvttosec1996(date)
    secondsSince12AM = secondsSince1996 - dayStartSecond

    dataFilePath = generate_data_file_path(date, structName)
    if (file_exists(dataFilePath) LT 1) then begin
	    ; file does not exist
	; print, 'WARNING: File ', dataFilePath, ' does not exist.'
	; print, 'No data that day?  Checking previous days events.'
	index = -1
	goto, OUT5
    endif

    structCount = get_files_struct_count(dataFilePath, structToUse)
    index = fix(structCount * (secondsSince12AM / 86400.))	; init guess

    get_lun, lun
    openr, lun, dataFilePath

    fileAssoc = assoc(lun, structToUse, HEADER_LENGTH)

    struct = fileAssoc[index]
    if (ct OR de) then stamp = struct.Second1996MinorFrame0
    ; print, '1: ', stamp, index
	; step forward past zero structs
    while (stamp EQ 0) do begin
	index = index + 1
	if (index EQ structCount) then goto, OUT1
	struct = fileAssoc[index]
	if (ct OR de) then stamp = struct.Second1996MinorFrame0
        ; print, '2: ', stamp, index
    endwhile

    OUT1:
    if ((stamp NE 0) AND (stamp LE secondsSince1996)) then begin
	    ; step forward past structs occuring before secondsSince1996
	while (stamp LE secondsSince1996) do begin
	    index = index + 1
	    if (index EQ structCount) then goto, OUT2
	    struct = fileAssoc[index]
	    if (ct OR de) then stamp = struct.Second1996MinorFrame0
            ; print, '3: ', stamp, index
	endwhile

	OUT2:
	    ; step back one struct
	index = index - 1
	struct = fileAssoc[index]
	if (ct OR de) then stamp = struct.Second1996MinorFrame0
	; print, '4: ', stamp, index
	    ; also step back past any zero structs
	while (stamp EQ 0) do begin
	    index = index - 1
	    if (index LT 0) then goto, OUT3
	    struct = fileAssoc[index]
	    if (ct OR de) then stamp = struct.Second1996MinorFrame0
	    ; print, '5: ', stamp, index
	endwhile

	OUT3:
    endif								       $
    else begin
	    ; step back before structs occuring later than secondsSince1996
	    ; also ignore zero time stamps
	while ((stamp GT secondsSince1996) OR (stamp EQ 0)) do begin
	    index = index - 1
	    if (index LT 0) then goto, OUT4
	    struct = fileAssoc[index]
	    if (ct OR de) then stamp = struct.Second1996MinorFrame0
	    ; print, '6: ', stamp, index
	endwhile

	OUT4:
    endelse

    OUT5:

    if (index GE 0) then begin
	    ; prior non-zero struct lies in this day's structs (thank god)
	if (ct OR de) then begin
	    struct = fileAssoc[index]
	    beginAndEndTimeTags[0] = struct.Second1996MinorFrame0
	    beginAndEndTimeTags[1] = struct.Second1996MinorFrame1
	endif
	close, lun
	free_lun, lun
	return, beginAndEndTimeTags
    endif								       $
    else begin
	    ; ugh! deal with looking in the previous day's data file...

	    ; determine path to previous day's data file
	day = day - 1
	date = get_date_from_day_1996(day)
	previousDaysPath = generate_data_file_path(date, structName)

	if (file_exists(previousDaysPath) LT 1) then begin
		; file does not exist.  return null times
	    ; print, 'WARNING: File ', previousDaysPath, ' does not exist.'
	    ; print, 'No data that day?'
	    beginAndEndTimeTags[0] = 0
	    beginAndEndTimeTags[1] = 0
	    close, lun
	    free_lun, lun
	    return, beginAndEndTimeTags
	endif

	    ; look for last non-zero struct in previous day's file
	structCount = get_files_struct_count(previousDaysPath, structToUse)
	get_lun, lun
	openr, lun, previousDaysPath
	fileAssoc = assoc(lun, structToUse, HEADER_LENGTH)
	index = structCount - 1
	struct = fileAssoc[index]
	if (ct OR de) then stamp = struct.Second1996MinorFrame0
	; print, '7: ', stamp, index
	    ; step back past zero structs
	while (stamp EQ 0) do begin
	    index = index - 1
	    if (index LT 0) then goto, OUT6
	    struct = fileAssoc[index]
	    if (ct OR de) then stamp = struct.Second1996MinorFrame0
	    ; print, '8: ', stamp, index
	endwhile

	OUT6:
	if (index GE 0) then begin
		; found a non-zero struct
	    if (ct OR de) then begin
		struct = fileAssoc[index]
		beginAndEndTimeTags[0] = struct.Second1996MinorFrame0
		beginAndEndTimeTags[1] = struct.Second1996MinorFrame1
	    endif
	    close, lun
	    free_lun, lun
	    return, beginAndEndTimeTags
	endif								       $
	else begin
		; did not find a non-zero struct
	    beginAndEndTimeTags[0] = 0
	    beginAndEndTimeTags[1] = 0
	    close, lun
	    free_lun, lun
	    return, beginAndEndTimeTags
	endelse

	close, lun
	free_lun, lun
    endelse
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MAIN FUNCTION get_structs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

function get_structs, secondsSince1996Start, secondsSince1996Stop, structToUse
    structName = strlowcase(tag_names(structToUse, /STRUCTURE_NAME))

    dataPeriod = get_data_type_period_from_struct_name(structName)
    if (dataPeriod GT 0) then begin
	getStampsAfter = secondsSince1996Start - dataPeriod
    endif								       $
    else begin
	startAndStop = get_begin_end_secs_of_struct_prior_to(		       $
					secondsSince1996Start, structToUse)
	if ((startAndStop[0] LE secondsSince1996Start) AND		       $
		(startAndStop[1] GE secondsSince1996Start)) then	       $
				getStampsAfter = startAndStop[0] - 1	       $
	else getStampsAfter = startAndStop[1]
    endelse
    getStampsUpTo = secondsSince1996Stop

    instrumentID = get_instrument_id_from_struct_name(structName)
    dataTypeSuffix = get_data_file_suffix_from_struct_name(structName)
    firstDayOfData = get_day_1996_from_date(				       $
				get_first_date_of_archived_data(structName))
    lastDayOfData = get_day_1996_from_date(				       $
				get_last_date_of_archived_data(structName))

    day1996Start = get_day_1996_from_seconds_since_1996(getStampsAfter)
    if (day1996Start LT firstDayOfData) then begin
	day1996Start = firstDayOfData
	getStampsAfter = get_seconds_since_1996_from_day_1996(day1996Start) - 1
    endif
    day1996Stop = get_day_1996_from_seconds_since_1996(getStampsUpTo)
    if (day1996Stop GT lastDayOfData) then begin
	day1996Stop = lastDayOfData
	getStampsUpTo = get_seconds_since_1996_from_day_1996(day1996Stop + 1)
    endif

    ;print, 'day1996Start = ', day1996Start
    ;print, 'day1996Stop = ', day1996Stop

    if (day1996Start EQ day1996Stop) then begin
	    ; easy case.  start and stop in the same day.

	    ; get_subset_of_days_structs will already terminate array with
	    ; a zeroed struct.  structArray can be returned directly.
	structArray = get_subset_of_days_structs(day1996Start, structToUse,    $
						getStampsAfter, getStampsUpTo)

	    ; if array holds more than 1 non-zero struct, chop off the
	    ; last non-zero struct if it is valid at getStampsUpTo.
	    ; this is to prevent the return of overlapping data on a
	    ; series of calls to getStructs.
	if (n_elements(structArray) GT 2) then				       $
	    dummy = zero_last_structs_if_valid_at_stop_time(structArray,       $
					getStampsUpTo, structToUse)

	return, structArray
    endif

    ; need to concatenate struct arrays from multiple days

	; deal with partial start day
    structArray = get_subset_of_days_structs(day1996Start, structToUse,	       $
	getStampsAfter, get_seconds_since_1996_from_day_1996(day1996Start + 1))
    n = n_elements(structArray)
    if (n GT 1) then fullArray = structArray[0:n - 2]	; remove zeroed struct

	; deal with any full days between start and end days
    for day = day1996Start + 1, day1996Stop - 1 do begin
	structArray = get_subset_of_days_structs(day, structToUse,	       $
				get_seconds_since_1996_from_day_1996(day) - 1, $
				get_seconds_since_1996_from_day_1996(day + 1))

	n = n_elements(structArray)
	if (n GT 1) then begin
	    structArray = structArray(0:n - 2)	; remove zeroed struct
	    if (n_elements(fullArray) GT 0) then		$
		fullArray = [fullArray, structArray]		$
	    else fullArray = structArray
	endif
    endfor

	; deal with partial last day
	; last thing to append.  already has zeroed struct at end.
    structArray = get_subset_of_days_structs(day1996Stop, structToUse,	       $
	get_seconds_since_1996_from_day_1996(day1996Stop) - 1, getStampsUpTo)
    if (n_elements(fullArray) GT 0) then fullArray = [fullArray, structArray]  $
    else fullArray = structArray
    
	    ; if array holds more than 1 non-zero struct, chop off the
	    ; last non-zero struct if it is valid at getStampsUpTo.
	    ; this is to prevent the return of overlapping data on a
	    ; series of calls to getStructs.
    if (n_elements(fullArray) GT 2) then				       $
	dummy = zero_last_structs_if_valid_at_stop_time(fullArray,	       $
					getStampsUpTo, structToUse)

    return, fullArray
end
